
# synopsis:
#   Search XML documents for regex pattern.
#   Use a tree walk.
# usage:
#   ./search_xml.exs [opts] regex-pattern file1 file2 ...
# options:
#   --verbose       Print extra information.  Default: false.
#   --fields="xxx"  Search in fields.  Possible fields are: tag, text,
#                   attr-name, attr-value.  Example: "tag:text:attr-value".
#                   Default: "".
# examples:
#   ./search_xml.exs --verbose pattern file1 file2 ...
#   ./search_xml.exs --verbose --fields="tag:text" "a-name" file1 file2 ...


defmodule SearchXml do

  #require IEx
  require Record
  Record.defrecord :xmlElement,
    Record.extract(:xmlElement, from_lib: "xmerl/include/xmerl.hrl")
  Record.defrecord :xmlText,
    Record.extract(:xmlText, from_lib: "xmerl/include/xmerl.hrl")
  Record.defrecord :xmlAttribute,
    Record.extract(:xmlAttribute, from_lib: "xmerl/include/xmerl.hrl")

  # Return true if pattern is found in field.  Else return false.
  @spec contains_pat(Keyword.t, atom, Regex.t,
                     String.t, String.t,
                     List.t(String.t), integer) :: boolean
  def contains_pat(
    fields, target_field,
    re_pattern, text,
    infilename, path,
    pos) do
    case Keyword.get(fields, target_field) do
      true ->
        case Regex.match?(re_pattern, text) do
          true ->
            path1 = Enum.map(path, fn(y) ->
              :io_lib.format("~s(~s)", [
                Atom.to_string(elem(y, 0)),
                Integer.to_string(elem(y,1))
              ])
            end)
            path2 = Enum.join(Enum.reverse(path1), "/")
            #IEx.pry
            field = to_string(target_field)
            IO.puts("#{infilename}[/#{path2}:#{field}(#{pos})]#{text}")
          _ -> nil
        end # case
      _ -> false
    end # case
  end # end contains_pat

  # Walk the XML tree.
  # Search fields for pattern.
  @spec walk(Tuple.t, Regex.t, List.t(String.t), String.t, List.t) :: nil
  defp walk(element, re_pattern, fields, infilename, parent_path) do
    case elem(element, 0) do
      :xmlElement ->
        tag = xmlElement(element, :name)
        pos = xmlElement(element, :pos)
        path = xmlElement(element, :parents)
        #:io.fwrite("** 1. path: ~p~n", [path])
        contains_pat(
          fields, :tag,
          re_pattern, to_string(tag),
          infilename, path, pos)
        attributes = xmlElement(element, :attributes)
        fn_show_attrib = fn (attrib) ->
          value = xmlAttribute(attrib, :value)
          pos1 = xmlAttribute(attrib, :pos)
          #:io.fwrite("** 2. path: ~p~n", [path])
          contains_pat(
            fields, :attribute,
            re_pattern, to_string(value),
            infilename, path, pos1)
        end # fn
        Enum.each(attributes, fn_show_attrib)
        fn_walk_children = fn (elm) ->
          walk(elm, re_pattern, fields, infilename, path)
        end # fn
        children = xmlElement(element, :content)
        Enum.each(children, fn_walk_children)
      :xmlText ->
        text = String.trim(List.to_string(xmlText(element, :value)))
        pos = xmlText(element, :pos)
        path = parent_path
        # If the text is an empty string, skip it.
        unless text == [] do
          #:io.fwrite("** 3. path: ~p~n", [path])
          contains_pat(fields, :text, re_pattern, text, infilename, path, pos)
        end # unless
      _ -> nil
    end # case
  end # end walk

  @doc """
  Search each of a list of files for pattern.
  The options argument contains the fields to be searched.
  """
  @spec search(Keyword.t, String.t, List.t(String.t)) :: nil
  def search(options, pattern, infilenames) do
    #IO.inspect(options)
    _verbose = Keyword.get(options, :verbose, false)
    fieldstrs = Keyword.get(options, :fields, "")
    field_fn = fn (val) -> {String.to_atom(val), true} end
    fields = Enum.map(String.split(fieldstrs, ":"), field_fn)
    #IO.inspect(fields)
    {:ok, re_pattern} = Regex.compile(pattern)
    fn_walk = fn (infilename) ->
      #sep = "--------------------------"
      #IO.puts("#{sep}\nDoc: #{infilename}\n#{sep}")
      {root, []} = :xmerl_scan.file(infilename)
      walk(root, re_pattern, fields, infilename, [])
    end # fn
    Enum.each(infilenames, fn_walk)
  end # end search

  # Print a usage message.
  # If variable `invalid` is not empty, print an error message.
  @spec usage(List.t(Tuple.t)) :: nil
  defp usage(invalid) do
    usage_message = """
    synopsis:
      Search XML documents for regex pattern.
    usage:
      ./search_xml.exs [opts] regex-pattern file1 file2 ...
    options:
      --verbose       Print extra information.  Default: false.
      --fields="xxx"  Search in fields.  Possible fields are: tag, text,
                      attr-name, attr-value.  Example: "tag:text:attr-value".
                      Default: "".
    examples:
      ./search_xml.exs --verbose pattern file1 file2 ...
      ./search_xml.exs --verbose --fields="tag:text" "a-name" file1 file2 ...
    """
    IO.puts(:stderr, usage_message)
    unless invalid == [] do
        IO.puts(:stderr, "invalid options:")
        IO.inspect(:stderr, invalid, [])
    end # unless
  end # usage

  @doc """
  Parse command line, then start the search.
  If the command line has an error or "--help", display usage.
  """
  @spec main() :: nil
  def main() do
    opt_spec = [help: :boolean, verbose: :boolean, fields: :list]
    {options, args, invalid} = OptionParser.parse(System.argv(), opt_spec)
    case invalid do
      [] ->
        help = Keyword.get(options, :help, false)
        case help do
          true ->
            usage([])
          false ->
            [pattern | infilenames] = args
            search(options, pattern, infilenames)
        end
      _ ->
        usage(invalid)
    end
  end # main

  def test() do
    opt = [fields: "tag:text:attribute"]
    pattern = "(Sammy)|(a00)|(artist)|(01)"
    files = ["albums01.xml", "albums02.xml"]
    search(opt, pattern, files)
  end # test

end # module SearchXml
